@@ -2,12 +2,7 @@
 # SNMP engine might be configured remotely (through SNMP).
 from pyasn1.compat.octets import null
 from pysnmp.carrier.asynsock import dispatch
-from pysnmp.carrier.asynsock.dgram import udp, udp6
-try:
-    from pysnmp.carrier.asynsock.dgram import unix
-    snmpLocalDomain = unix.snmpLocalDomain
-except ImportError: # UNIX-specific -- may not be always available
-    pass
+from pysnmp.carrier.asynsock.dgram import udp, udp6, unix
 from pysnmp.proto import rfc3412
 from pysnmp.entity import engine
 from pysnmp.proto.secmod.rfc3414.auth import hmacmd5, hmacsha, noauth
@@ -22,6 +17,7 @@ from pysnmp import error
 # Transports
 snmpUDPDomain = udp.snmpUDPDomain
 snmpUDP6Domain = udp6.snmpUDP6Domain
+snmpLocalDomain = unix.snmpLocalDomain
 
 # Auth protocol
 usmHMACMD5AuthProtocol = hmacmd5.HmacMd5.serviceID
@@ -269,8 +265,7 @@ def addTargetAddr(
         (snmpTargetAddrEntry.name + (3,) + tblIdx, transportAddress),
         (snmpTargetAddrEntry.name + (4,) + tblIdx, timeout),
         (snmpTargetAddrEntry.name + (5,) + tblIdx, retryCount),
-         # XXX
-        (snmpTargetAddrEntry.name + (6,) + tblIdx,'%s %s'%(addrName,tagList)),
+        (snmpTargetAddrEntry.name + (6,) + tblIdx, tagList),
         (snmpTargetAddrEntry.name + (7,) + tblIdx, params),)
         )
 
@@ -294,9 +289,7 @@ def delSocketTransport(snmpEngine, transportDomain):
     """Unregister transport object at socket dispatcher of snmpEngine"""
     if not snmpEngine.transportDispatcher:
         return
-    snmpEngine.transportDispatcher.unregisterTransport(
-        transportDomain
-        )
+    snmpEngine.transportDispatcher.unregisterTransport(transportDomain)
     snmpEngine.unregisterTransportDispatcher()
 
 # VACM shortcuts
