@@ -4,7 +4,6 @@ from pysnmp.entity.rfc3413 import config
 from pysnmp.proto.proxy import rfc2576
 from pysnmp import error, nextid, debug
 from pyasn1.type import univ
-from pyasn1.compat.octets import null
 
 getNextHandle = nextid.Integer(0x7fffffff)
                              
@@ -23,7 +22,9 @@ def getNextVarBinds(origVarBinds, varBinds):
     rspVarBinds = []
     while idx:
         idx = idx - 1
-        if isinstance(varBinds[idx][1], univ.Null):
+        if varBinds[idx][1].tagSet in (rfc1905.NoSuchObject.tagSet,
+                                       rfc1905.NoSuchInstance.tagSet,
+                                       rfc1905.EndOfMibView.tagSet):
             nonNulls = nonNulls - 1
         elif origVarBinds[idx][0].asTuple() >= varBinds[idx][0].asTuple():
             errorIndication = errind.oidNotIncreasing
@@ -39,6 +40,7 @@ class CommandGeneratorBase:
     _null = univ.Null('')
     def __init__(self):
         self.__pendingReqs = {}
+        self.__SnmpEngineID, self.__SnmpAdminString = None, None
             
     def processResponsePdu(
         self,
@@ -57,6 +59,9 @@ class CommandGeneratorBase:
         ):
         (cbFun, cbCtx) = cbInfo
         # 3.1.1
+        if sendPduHandle not in self.__pendingReqs:
+            raise error.PySnmpError('Missing sendPduHandle %s' % sendPduHandle)
+
         ( origTransportDomain,
           origTransportAddress,
           origMessageProcessingModel,
@@ -72,6 +77,7 @@ class CommandGeneratorBase:
           origRetries,
           origSendRequestHandle
           ) = self.__pendingReqs[sendPduHandle]
+  
         del self.__pendingReqs[sendPduHandle]
 
         snmpEngine.transportDispatcher.jobFinished(id(self))
@@ -79,7 +85,14 @@ class CommandGeneratorBase:
         # 3.1.3
         if statusInformation:
             debug.logger & debug.flagApp and debug.logger('processResponsePdu: sendPduHandle %s, statusInformation %s' % (sendPduHandle, statusInformation))
-            if origRetries == origRetryCount:
+            errorIndication = statusInformation['errorIndication']
+            # SNMP engine discovery will take extra retries, allow that
+            if errorIndication in (errind.notInTimeWindow,
+                                   errind.unknownEngineID) and \
+                                   origRetries == origRetryCount + 2 or \
+               errorIndication not in (errind.notInTimeWindow,
+                                       errind.unknownEngineID) and \
+                                   origRetries == origRetryCount:
                 debug.logger & debug.flagApp and debug.logger('processResponsePdu: sendPduHandle %s, retry count %d exceeded' % (sendPduHandle, origRetries))
                 cbFun(origSendRequestHandle,
                       statusInformation['errorIndication'], 0, 0, (),
@@ -156,7 +169,7 @@ class CommandGeneratorBase:
         cbFun,
         cbCtx=None,
         contextEngineId=None,
-        contextName=null
+        contextName=''
         ):
         raise error.PySnmpError('Method not implemented')
 
@@ -180,6 +193,18 @@ class CommandGeneratorBase:
         cbInfo
         ):
         (processResponsePdu, cbCtx) = cbInfo
+
+        # Convert timeout in seconds into timeout in timer ticks
+        timeoutInTicks = float(timeout)/100/snmpEngine.transportDispatcher.getTimerResolution()
+
+        if not self.__SnmpEngineID or not self.__SnmpAdminString:
+            self.__SnmpEngineID, self.__SnmpAdminString = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('SNMP-FRAMEWORK-MIB', 'SnmpEngineID', 'SnmpAdminString')
+
+        # Cast possible strings into bytes
+        if contextEngineId:
+            contextEngineId = self.__SnmpEngineID(contextEngineId)
+        contextName = self.__SnmpAdminString(contextName)
+
         # 3.1
         sendPduHandle = snmpEngine.msgAndPduDsp.sendPdu(
             snmpEngine,
@@ -193,8 +218,8 @@ class CommandGeneratorBase:
             contextName,
             pduVersion,
             reqPDU,
-            1,                                 # expectResponse
-            float(timeout)/100 + time.time(),  # timeout
+            1,                          # expectResponse
+            timeoutInTicks,
             processResponsePdu,
             cbCtx
             )
@@ -220,6 +245,34 @@ class CommandGeneratorBase:
             sendRequestHandle,
             )
 
+    def _handleResponse(
+        self,
+        snmpEngine,
+        transportDomain,
+        transportAddress,
+        messageProcessingModel,
+        securityModel,
+        securityName,
+        securityLevel,
+        contextEngineId,
+        contextName,
+        pduVersion,
+        PDU,
+        timeout,
+        retryCount,
+        pMod,
+        rspPDU,
+        sendRequestHandle,
+        cbInfo
+        ):
+        (cbFun, cbCtx) = cbInfo        
+        cbFun(sendRequestHandle,
+              None,
+              pMod.apiPDU.getErrorStatus(rspPDU),
+              pMod.apiPDU.getErrorIndex(rspPDU),
+              pMod.apiPDU.getVarBinds(rspPDU),
+              cbCtx)
+
 class GetCommandGenerator(CommandGeneratorBase):
     def sendReq(
         self,
@@ -229,7 +282,7 @@ class GetCommandGenerator(CommandGeneratorBase):
         cbFun,
         cbCtx=None,
         contextEngineId=None,
-        contextName=null
+        contextName=''
         ):
         ( transportDomain,
           transportAddress,
@@ -270,34 +323,6 @@ class GetCommandGenerator(CommandGeneratorBase):
 
         return requestHandle
     
-    def _handleResponse(
-        self,
-        snmpEngine,
-        transportDomain,
-        transportAddress,
-        messageProcessingModel,
-        securityModel,
-        securityName,
-        securityLevel,
-        contextEngineId,
-        contextName,
-        pduVersion,
-        PDU,
-        timeout,
-        retryCount,
-        pMod,
-        rspPDU,
-        sendRequestHandle,
-        cbInfo
-        ):
-        (cbFun, cbCtx) = cbInfo        
-        cbFun(sendRequestHandle,
-              None,
-              pMod.apiPDU.getErrorStatus(rspPDU),
-              pMod.apiPDU.getErrorIndex(rspPDU),
-              pMod.apiPDU.getVarBinds(rspPDU),
-              cbCtx)
-
 class SetCommandGenerator(CommandGeneratorBase):
     def sendReq(
         self,
@@ -307,7 +332,7 @@ class SetCommandGenerator(CommandGeneratorBase):
         cbFun,
         cbCtx=None,
         contextEngineId=None,
-        contextName=null
+        contextName=''
         ):
         ( transportDomain,
           transportAddress,
@@ -353,35 +378,7 @@ class SetCommandGenerator(CommandGeneratorBase):
 
         return requestHandle
 
-    def _handleResponse(
-        self,
-        snmpEngine,
-        transportDomain,
-        transportAddress,
-        messageProcessingModel,
-        securityModel,
-        securityName,
-        securityLevel,
-        contextEngineId,
-        contextName,
-        pduVersion,
-        PDU,
-        timeout,
-        retryCount,
-        pMod,
-        rspPDU,
-        sendRequestHandle,
-        cbInfo
-        ):
-        (cbFun, cbCtx) = cbInfo        
-        cbFun(sendRequestHandle,
-              None,
-              pMod.apiPDU.getErrorStatus(rspPDU),
-              pMod.apiPDU.getErrorIndex(rspPDU),
-              pMod.apiPDU.getVarBinds(rspPDU),
-              cbCtx)
-
-class NextCommandGenerator(CommandGeneratorBase):
+class NextCommandGeneratorSingleRun(CommandGeneratorBase):
     def sendReq(
         self,
         snmpEngine,
@@ -390,7 +387,7 @@ class NextCommandGenerator(CommandGeneratorBase):
         cbFun,
         cbCtx=None,
         contextEngineId=None,
-        contextName=null
+        contextName=''
         ):
         ( transportDomain,
           transportAddress,
@@ -431,6 +428,7 @@ class NextCommandGenerator(CommandGeneratorBase):
 
         return requestHandle
     
+class NextCommandGenerator(NextCommandGeneratorSingleRun):
     def _handleResponse(
         self,
         snmpEngine,
@@ -452,12 +450,13 @@ class NextCommandGenerator(CommandGeneratorBase):
         cbInfo
         ):
         (cbFun, cbCtx) = cbInfo
+
         varBindTable = pMod.apiPDU.getVarBindTable(PDU, rspPDU)
 
         if pMod.apiPDU.getErrorStatus(rspPDU):
-            errorIndication = None
+            errorIndication, varBinds = None, ()
         elif not varBindTable:
-            errorIndication = errind.emptyResponse
+            errorIndication, varBinds = errind.emptyResponse, ()
         else:
             errorIndication, varBinds = getNextVarBinds(
                 pMod.apiPDU.getVarBinds(PDU), varBindTable[-1]
@@ -495,7 +494,7 @@ class NextCommandGenerator(CommandGeneratorBase):
             (self.processResponsePdu, (cbFun, cbCtx))            
             )
 
-class BulkCommandGenerator(CommandGeneratorBase):
+class BulkCommandGeneratorSingleRun(CommandGeneratorBase):
     def sendReq(
         self,
         snmpEngine,
@@ -506,7 +505,7 @@ class BulkCommandGenerator(CommandGeneratorBase):
         cbFun,
         cbCtx=None,
         contextEngineId=None,
-        contextName=null
+        contextName=''
         ):
         ( transportDomain,
           transportAddress,
@@ -552,6 +551,7 @@ class BulkCommandGenerator(CommandGeneratorBase):
 
         return requestHandle
     
+class BulkCommandGenerator(BulkCommandGeneratorSingleRun):
     def _handleResponse(
         self,
         snmpEngine,
@@ -576,9 +576,9 @@ class BulkCommandGenerator(CommandGeneratorBase):
         varBindTable = pMod.apiBulkPDU.getVarBindTable(PDU, rspPDU)
 
         if pMod.apiBulkPDU.getErrorStatus(rspPDU):
-            errorIndication = None
+            errorIndication, varBinds = None, ()
         elif not varBindTable:
-            errorIndication = errind.emptyResponse
+            errorIndication, varBinds = errind.emptyResponse, ()
         else:
             errorIndication, varBinds = getNextVarBinds(
                 pMod.apiBulkPDU.getVarBinds(PDU), varBindTable[-1]
