@@ -1,8 +1,10 @@
-import socket, sys
 from pysnmp.entity import engine, config
-from pysnmp.entity.rfc3413 import cmdgen, mibvar
-from pysnmp.carrier.asynsock.dgram import udp
-from pysnmp.proto import errind
+from pysnmp.entity.rfc3413 import cmdgen
+from pysnmp.entity.rfc3413.oneliner.mibvar import MibVariable
+from pysnmp.entity.rfc3413.oneliner.auth import CommunityData, UsmUserData
+from pysnmp.entity.rfc3413.oneliner.target import UdpTransportTarget, \
+    Udp6TransportTarget, UnixTransportTarget 
+from pysnmp.proto import rfc1905, errind
 from pysnmp.smi import view
 from pysnmp import nextid, error
 from pyasn1.type import univ, base
@@ -23,133 +25,11 @@ usmNoPrivProtocol = config.usmNoPrivProtocol
 
 nextID = nextid.Integer(0xffffffff)
 
-class CommunityData:
-    mpModel = 1 # Default is SMIv2
-    securityModel = mpModel+1
-    securityLevel = 'noAuthNoPriv'
-    contextName = null
-    def __init__(self, securityName, communityName, mpModel=None,
-                 contextEngineId=None, contextName=None):
-        self.securityName = securityName
-        self.communityName = communityName
-        if mpModel is not None:
-            self.mpModel = mpModel
-            self.securityModel = mpModel + 1
-        self.contextEngineId = contextEngineId
-        if contextName is not None:
-            self.contextName = contextName
-        self.__cmp = self.mpModel, self.securityModel, self.securityLevel, self.securityName, self.communityName, self.contextEngineId, self.contextName
-        self.__hash = hash(self.__cmp)
-            
-    def __repr__(self):
-        return '%s("%s", <COMMUNITY>, %s, %s, %s)' % (
-            self.__class__.__name__,
-            self.securityName,
-            self.mpModel,
-            self.contextEngineId,
-            self.contextName
-            )
-
-    def __hash__(self): return self.__hash
-
-    def __eq__(self, other): return self.__cmp == other
-    def __ne__(self, other): return self.__cmp != other
-    def __lt__(self, other): return self.__cmp < other
-    def __le__(self, other): return self.__cmp <= other
-    def __gt__(self, other): return self.__cmp > other
-    def __ge__(self, other): return self.__cmp >= other
-
-class UsmUserData:
-    authKey = privKey = None
-    authProtocol = usmNoAuthProtocol
-    privProtocol = usmNoPrivProtocol
-    securityLevel = 'noAuthNoPriv'
-    securityModel = 3
-    mpModel = 2
-    contextName = null
-    def __init__(self, securityName,
-                 authKey=None, privKey=None,
-                 authProtocol=None, privProtocol=None,
-                 contextEngineId=None, contextName=None):
-        self.securityName = securityName
-        
-        if authKey is not None:
-            self.authKey = authKey
-            if authProtocol is None:
-                self.authProtocol = usmHMACMD5AuthProtocol
-            else:
-                self.authProtocol = authProtocol
-            if self.securityLevel != 'authPriv':
-                self.securityLevel = 'authNoPriv'
-
-        if privKey is not None:
-            self.privKey = privKey
-            if self.authProtocol == usmNoAuthProtocol:
-                raise error.PySnmpError('Privacy implies authenticity')
-            self.securityLevel = 'authPriv'
-            if privProtocol is None:
-                self.privProtocol = usmDESPrivProtocol
-            else:
-                self.privProtocol = privProtocol
-
-        self.contextEngineId = contextEngineId
-        if contextName is not None:
-            self.contextName = contextName
-        
-        self.__cmp = self.mpModel, self.securityModel, self.securityLevel, self.securityName, self.authProtocol, self.authKey, self.privProtocol, self.privKey, self.contextEngineId
-        self.__hash = hash(self.__cmp)
-
-    def __repr__(self):
-        return '%s("%s", <AUTHKEY>, <PRIVKEY>, %s, %s, %s, %s)' % (
-            self.__class__.__name__,
-            self.securityName,
-            self.authProtocol,
-            self.privProtocol,
-            self.contextEngineId,
-            self.contextName
-            )
-
-    def __hash__(self): return self.__hash
-    
-    def __eq__(self, other): return self.__cmp == other
-    def __ne__(self, other): return self.__cmp != other
-    def __lt__(self, other): return self.__cmp < other
-    def __le__(self, other): return self.__cmp <= other
-    def __gt__(self, other): return self.__cmp > other
-    def __ge__(self, other): return self.__cmp >= other
-    
-class UdpTransportTarget:
-    transportDomain = udp.domainName
-    def __init__(self, transportAddr, timeout=1, retries=5):
-        self.transportAddr = (
-            socket.gethostbyname(transportAddr[0]), transportAddr[1]
-            )
-        self.timeout = timeout
-        self.retries = retries
-
-    def __repr__(self): return '%s(("%s", %s), %s, %s)' % (
-        self.__class__.__name__,
-        self.transportAddr[0], self.transportAddr[1],
-        self.timeout, self.retries
-        )
-
-    def __hash__(self): return hash(self.transportAddr)
-    
-    def __eq__(self, other): return self.transportAddr == other
-    def __ne__(self, other): return self.transportAddr != other
-    def __lt__(self, other): return self.transportAddr < other
-    def __le__(self, other): return self.transportAddr <= other
-    def __gt__(self, other): return self.transportAddr > other
-    def __ge__(self, other): return self.transportAddr >= other
-    
-    def openClientMode(self):
-        self.transport = udp.UdpSocketTransport().openClientMode()
-        return self.transport
-        
 class AsynCommandGenerator:
     _null = univ.Null('')
     def __init__(self, snmpEngine=None):
         self.__knownAuths = {}
+        self.__knownParams = {}
         self.__knownTransports = {}
         self.__knownTransportAddrs = {}
         if snmpEngine is None:
@@ -158,17 +38,12 @@ class AsynCommandGenerator:
             self.snmpEngine = snmpEngine
         self.mibViewController = view.MibViewController(
             self.snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder
-            )
+        )
 
     def __del__(self): self.uncfgCmdGen()
 
-    def cfgCmdGen(self, authData, transportTarget, tagList=null):
-        if isinstance(authData, CommunityData):
-            tagList = '%s %s' % (tagList, authData.securityName)
-        if authData in self.__knownAuths:
-            paramsName = self.__knownAuths[authData]
-        else:
-            paramsName = 'p%s' % nextID()
+    def cfgCmdGen(self, authData, transportTarget):
+        if authData.securityName not in self.__knownAuths:
             if isinstance(authData, CommunityData):
                 config.addV1System(
                     self.snmpEngine,
@@ -176,13 +51,8 @@ class AsynCommandGenerator:
                     authData.communityName,
                     authData.contextEngineId,
                     authData.contextName,
-                    tagList
-                    )
-                config.addTargetParams(
-                    self.snmpEngine, paramsName,
-                    authData.securityName, authData.securityLevel,
-                    authData.mpModel
-                    )
+                    authData.tag
+                )
             elif isinstance(authData, UsmUserData):
                 config.addV3User(
                     self.snmpEngine,
@@ -190,14 +60,22 @@ class AsynCommandGenerator:
                     authData.authProtocol, authData.authKey,
                     authData.privProtocol, authData.privKey,
                     authData.contextEngineId
-                    )
-                config.addTargetParams(
-                    self.snmpEngine, paramsName,
-                    authData.securityName, authData.securityLevel
-                    )
+                )
             else:
                 raise error.PySnmpError('Unsupported authentication object')
-            self.__knownAuths[authData] = paramsName
+
+            self.__knownAuths[authData.securityName] = authData
+
+        k = authData.securityName, authData.securityLevel, authData.mpModel
+        if k in self.__knownParams:
+            paramsName = self.__knownParams[k]
+        else:
+            paramsName = 'p%s' % nextID()
+            config.addTargetParams(
+                self.snmpEngine, paramsName,
+                authData.securityName, authData.securityLevel, authData.mpModel
+            )
+            self.__knownParams[k] = paramsName
 
         if transportTarget.transportDomain not in self.__knownTransports:
             transport = transportTarget.openClientMode()
@@ -205,10 +83,11 @@ class AsynCommandGenerator:
                 self.snmpEngine,
                 transportTarget.transportDomain,
                 transport
-                )
+            )
             self.__knownTransports[transportTarget.transportDomain] = transport
 
-        k = paramsName, transportTarget, tagList
+        k = paramsName, transportTarget, transportTarget.tagList
+
         if k in self.__knownTransportAddrs:
             addrName = self.__knownTransportAddrs[k]
         else:
@@ -220,33 +99,33 @@ class AsynCommandGenerator:
                 paramsName,
                 transportTarget.timeout * 100,
                 transportTarget.retries,
-                tagList                
-                )
+                transportTarget.tagList
+            )
             self.__knownTransportAddrs[k] = addrName
 
         return addrName, paramsName
 
     def uncfgCmdGen(self):
-        for authData, paramsName in self.__knownAuths.items():
+        for authData in self.__knownAuths.values():
             if isinstance(authData, CommunityData):
                 config.delV1System(
                     self.snmpEngine,
                     authData.securityName
                     )
-                config.delTargetParams(
-                    self.snmpEngine, paramsName
-                    )
             elif isinstance(authData, UsmUserData):
                 config.delV3User(
                     self.snmpEngine, authData.securityName
                     )
-                config.delTargetParams(
-                    self.snmpEngine, paramsName
-                    )
             else:
                 raise error.PySnmpError('Unsupported authentication object')
         self.__knownAuths.clear()
 
+        for paramsName in self.__knownParams.values():
+            config.delTargetParams(
+                self.snmpEngine, paramsName
+                )
+        self.__knownParams.clear()
+        
         for transportDomain, transport in self.__knownTransports.items():
             config.delSocketTransport(
                 self.snmpEngine, transportDomain
@@ -260,99 +139,183 @@ class AsynCommandGenerator:
                 )
         self.__knownTransportAddrs.clear()
                 
-    if sys.version_info[0] <= 2:
-        intTypes = (int, long)
-    else:
-        intTypes = (int,)
-
     def makeReadVarBinds(self, varNames):
         varBinds = []
         for varName in varNames:
-            if isinstance(varName[0], self.intTypes):
-                name = varName
+            if isinstance(varName, MibVariable):
+                varName.resolveWithMib(
+                    self.mibViewController, oidOnly=True
+                )
+            elif isinstance(varName[0], tuple):  # legacy
+                varName = MibVariable(varName[0][0], varName[0][1], *varName[1:]).resolveWithMib(self.mibViewController)
             else:
-                name, oid = mibvar.mibNameToOid(
-                    self.mibViewController, varName
-                    )
-                name = name + oid
-            varBinds.append((name, self._null))
+                varName = MibVariable(varName).resolveWithMib(self.mibViewController, oidOnly=True)
+            varBinds.append((varName, self._null))
         return varBinds
 
+    def unmakeVarBinds(self, varBinds, lookupNames, lookupValues):
+        if lookupNames or lookupValues:
+            _varBinds = []
+            for name, value in varBinds:
+                varName = MibVariable(name).resolveWithMib(self.mibViewController)
+                if lookupNames:
+                    name = varName
+                if lookupValues:
+                    if value.tagSet not in (rfc1905.NoSuchObject.tagSet,
+                                            rfc1905.NoSuchInstance.tagSet,
+                                            rfc1905.EndOfMibView.tagSet):
+                        if varName.isFullyResolved():
+                            value = varName.getMibNode().getSyntax().clone(value)
+                _varBinds.append((name, value))
+            return _varBinds
+        else:
+            return varBinds
+
     # Async SNMP apps
 
-    def getCmd(self, authData, transportTarget, varNames, cbInfo):
+    def getCmd(self, authData, transportTarget, varNames, cbInfo,
+               lookupNames=False, lookupValues=False):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBinds, cbCtx):
+            lookupNames, lookupValues, cbFun, cbCtx = cbCtx
+            return cbFun(
+                sendRequestHandle,
+                errorIndication,
+                errorStatus,
+                errorIndex,
+                self.unmakeVarBinds(varBinds, lookupNames, lookupValues),
+                cbCtx
+            )
+
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
-            )
-        varBinds = self.makeReadVarBinds(varNames)
+        )
         return cmdgen.GetCommandGenerator().sendReq(
-            self.snmpEngine, addrName, varBinds, cbFun, cbCtx,
+            self.snmpEngine,
+            addrName,
+            self.makeReadVarBinds(varNames),
+            __cbFun,
+            (lookupNames, lookupValues, cbFun, cbCtx),
             authData.contextEngineId, authData.contextName
-            )
+        )
     
     asyncGetCmd = getCmd
     
-    def setCmd(self, authData, transportTarget, varBinds, cbInfo):
+    def setCmd(self, authData, transportTarget, varBinds, cbInfo,
+               lookupNames=False, lookupValues=False):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBinds, cbCtx):
+            lookupNames, lookupValues, cbFun, cbCtx = cbCtx
+            return cbFun(
+                sendRequestHandle,
+                errorIndication,
+                errorStatus,
+                errorIndex,
+                self.unmakeVarBinds(varBinds, lookupNames, lookupValues),
+                cbCtx
+            )
+
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
-            )
+        )
         __varBinds = []
         for varName, varVal in varBinds:
-            name, oid = mibvar.mibNameToOid(
-                self.mibViewController, varName
-                )
-            if not isinstance(varVal, base.AbstractSimpleAsn1Item):
-                ((symName, modName), suffix) = mibvar.oidToMibName(
-                    self.mibViewController, name + oid
-                    )
-                syntax = mibvar.cloneFromMibValue(
-                    self.mibViewController, modName, symName, varVal
-                    )
-                if syntax is None:
-                    raise error.PySnmpError(
-                        'Value type MIB lookup failed for %s' % repr(varName)
-                        )
-                varVal = syntax.clone(varVal)
-            __varBinds.append((name + oid, varVal))
+            if isinstance(varName, MibVariable):
+                varName.resolveWithMib(self.mibViewController)
+                if not isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
+            elif isinstance(varName[0], tuple):  # legacy
+                varName = MibVariable(varName[0][0], varName[0][1], *varName[1:]).resolveWithMib(self.mibViewController)
+                if not isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
+            else:
+                if isinstance(varVal, base.AbstractSimpleAsn1Item):
+                    varName = MibVariable(varName).resolveWithMib(self.mibViewController, oidOnly=True)
+                else:
+                    varName = MibVariable(varName).resolveWithMib(self.mibViewController)
+                    varVal = varName.getMibNode().getSyntax().clone(varVal)
+
+            __varBinds.append((varName, varVal))
         return cmdgen.SetCommandGenerator().sendReq(
-            self.snmpEngine, addrName, __varBinds, cbFun, cbCtx,
-            authData.contextEngineId, authData.contextName
-            )
+            self.snmpEngine,
+            addrName,
+            __varBinds,
+            __cbFun,
+            (lookupNames, lookupValues, cbFun, cbCtx),
+            authData.contextEngineId,
+            authData.contextName
+        )
     
     asyncSetCmd = setCmd
     
-    def nextCmd(self, authData, transportTarget, varNames, cbInfo):
+    def nextCmd(self, authData, transportTarget, varNames, cbInfo,
+                lookupNames=False, lookupValues=False):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBindTable, cbCtx):
+            lookupNames, lookupValues, cbFun, cbCtx = cbCtx
+            return cbFun(
+                sendRequestHandle,
+                errorIndication,
+                errorStatus,
+                errorIndex,
+                [ self.unmakeVarBinds(varBindTableRow, lookupNames, lookupValues) for varBindTableRow in varBindTable ],
+                cbCtx
+            )
+
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
-            )
-        varBinds = self.makeReadVarBinds(varNames)
+        )
         return cmdgen.NextCommandGenerator().sendReq(
-            self.snmpEngine, addrName, varBinds, cbFun, cbCtx,
+            self.snmpEngine,
+            addrName,
+            self.makeReadVarBinds(varNames),
+            __cbFun,
+            (lookupNames, lookupValues, cbFun, cbCtx),
             authData.contextEngineId, authData.contextName
-            )
+        )
 
     asyncNextCmd = nextCmd
     
-    def bulkCmd(self, authData, transportTarget, nonRepeaters, maxRepetitions,
-                varNames, cbInfo):
+    def bulkCmd(self, authData, transportTarget,
+                nonRepeaters, maxRepetitions, varNames, cbInfo,
+                lookupNames=False, lookupValues=False):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBindTable, cbCtx):
+            lookupNames, lookupValues, cbFun, cbCtx = cbCtx
+            return cbFun(
+                sendRequestHandle,
+                errorIndication,
+                errorStatus,
+                errorIndex,
+                [ self.unmakeVarBinds(varBindTableRow, lookupNames, lookupValues) for varBindTableRow in varBindTable ],
+                cbCtx
+            )
+
         (cbFun, cbCtx) = cbInfo
         addrName, paramsName = self.cfgCmdGen(
             authData, transportTarget
-            )
+        )
         varBinds = self.makeReadVarBinds(varNames)
         return cmdgen.BulkCommandGenerator().sendReq(
-            self.snmpEngine, addrName,
-            nonRepeaters, maxRepetitions, varBinds, cbFun, cbCtx,
+            self.snmpEngine,
+            addrName,
+            nonRepeaters, maxRepetitions,
+            varBinds,
+            __cbFun,
+            (lookupNames, lookupValues, cbFun, cbCtx),
             authData.contextEngineId, authData.contextName
-            )
+        )
 
     asyncBulkCmd = bulkCmd
 
 class CommandGenerator:
-    lexicographicMode = ignoreNonIncreasingOid = None
     def __init__(self, snmpEngine=None, asynCmdGen=None):
         if asynCmdGen is None:
             self.__asynCmdGen = AsynCommandGenerator(snmpEngine)
@@ -362,12 +325,11 @@ class CommandGenerator:
         # compatibility attributes
         self.snmpEngine = self.__asynCmdGen.snmpEngine
         self.mibViewController = self.__asynCmdGen.mibViewController
-        
-    def getCmd(self, authData, transportTarget, *varNames):
-        def __cbFun(
-            sendRequestHandle, errorIndication, errorStatus, errorIndex,
-            varBinds, appReturn
-            ):
+       
+    def getCmd(self, authData, transportTarget, *varNames, **kwargs):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBinds, appReturn):
             appReturn['errorIndication'] = errorIndication
             appReturn['errorStatus'] = errorStatus
             appReturn['errorIndex'] = errorIndex
@@ -375,46 +337,61 @@ class CommandGenerator:
 
         appReturn = {}
         self.__asynCmdGen.getCmd(
-            authData, transportTarget, varNames, (__cbFun, appReturn)
-            )
+            authData,
+            transportTarget,
+            varNames,
+            (__cbFun, appReturn),
+            kwargs.get('lookupNames', False),
+            kwargs.get('lookupValues', False)
+        )
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
         return (
             appReturn['errorIndication'],
             appReturn['errorStatus'],
             appReturn['errorIndex'],
             appReturn['varBinds']
-            )
+        )
 
-    def setCmd(self, authData, transportTarget, *varBinds):
-        def __cbFun(
-            sendRequestHandle, errorIndication, errorStatus, errorIndex,
-            varBinds, appReturn
-            ):
+    def setCmd(self, authData, transportTarget, *varBinds, **kwargs):
+        def __cbFun(sendRequestHandle,
+                    errorIndication, errorStatus, errorIndex,
+                    varBinds, appReturn):
             appReturn['errorIndication'] = errorIndication
             appReturn['errorStatus'] = errorStatus
             appReturn['errorIndex'] = errorIndex
             appReturn['varBinds'] = varBinds
 
+        lookupNames = kwargs.get('lookupNames', False)        
+        lookupValues = kwargs.get('lookupValues', False)
+
         appReturn = {}
         self.__asynCmdGen.setCmd(
-            authData, transportTarget, varBinds, (__cbFun, appReturn)
-            )
+            authData,
+            transportTarget,
+            varBinds,
+            (__cbFun, appReturn),
+            kwargs.get('lookupNames', False),
+            kwargs.get('lookupValues', False)
+        )
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
         return (
             appReturn['errorIndication'],
             appReturn['errorStatus'],
             appReturn['errorIndex'],
             appReturn['varBinds']
-            )
+        )
 
-    def nextCmd(self, authData, transportTarget, *varNames):
+    def nextCmd(self, authData, transportTarget, *varNames, **kwargs):
         def __cbFun(sendRequestHandle, errorIndication,
                     errorStatus, errorIndex, varBindTable, cbCtx):
             (self, varBindHead, varBindTotalTable, appReturn) = cbCtx
-            if errorStatus or \
-               errorIndication and not self.ignoreNonIncreasingOid or \
-               errorIndication and self.ignoreNonIncreasingOid and \
-               not isinstance(errorIndication, errind.OidNotIncreasing):
+            if (ignoreNonIncreasingOid or \
+                        hasattr(self, 'ignoreNonIncreasingOid') and \
+                        self.ignoreNonIncreasingOid ) and \
+                    errorIndication and \
+                    isinstance(errorIndication, errind.OidNotIncreasing):
+                errorIndication = None
+            if errorStatus or errorIndication:
                 appReturn['errorIndication'] = errorIndication
                 if errorStatus == 2:
                     # Hide SNMPv1 noSuchName error which leaks in here
@@ -427,78 +404,26 @@ class CommandGenerator:
                 appReturn['varBindTable'] = varBindTotalTable
                 return
             else:
-                varBindTableRow = varBindTable[-1]
+                if maxRows and len(varBindTotalTable) >= maxRows or \
+                        hasattr(self, 'maxRows') and self.maxRows and \
+                        len(varBindTotalTable) >= self.maxRows:
+                    appReturn['errorIndication'] = errorIndication
+                    appReturn['errorStatus'] = errorStatus
+                    appReturn['errorIndex'] = errorIndex
+                    if hasattr(self, 'maxRows'):
+                        appReturn['varBindTable'] = varBindTotalTable[:self.maxRows]
+                    else:
+                        appReturn['varBindTable'] = varBindTotalTable[:maxRows]
+                    return
+                
+                varBindTableRow = varBindTable and varBindTable[-1] or varBindTable
                 for idx in range(len(varBindTableRow)):
                     name, val = varBindTableRow[idx]
                     # XXX extra rows
                     if not isinstance(val, univ.Null):
-                        if self.lexicographicMode:
-                            if varBindHead[idx] <= name:
-                                break
-                        else:
-                            if varBindHead[idx].isPrefixOf(name):
-                                break
-                else:
-                    appReturn['errorIndication'] = errorIndication
-                    appReturn['errorStatus'] = errorStatus
-                    appReturn['errorIndex'] = errorIndex
-                    appReturn['varBindTable'] = varBindTotalTable
-                    return
-                varBindTotalTable.extend(varBindTable)
-
-            return 1 # continue table retrieval
-
-        varBindHead = []
-        for varName in varNames:
-            name, suffix = mibvar.mibNameToOid(
-                self.__asynCmdGen.mibViewController, varName
-                )
-            varBindHead.append(univ.ObjectIdentifier(name + suffix))
-
-        appReturn = {}
-        self.__asynCmdGen.nextCmd(
-            authData, transportTarget, varNames,
-            (__cbFun, (self, varBindHead,[],appReturn))
-            )
-
-        self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
-
-        return (
-            appReturn['errorIndication'],
-            appReturn['errorStatus'],
-            appReturn['errorIndex'],
-            appReturn['varBindTable']
-            )
-
-    def bulkCmd(self, authData, transportTarget,
-                nonRepeaters, maxRepetitions, *varNames):
-        def __cbFun(sendRequestHandle, errorIndication,
-                    errorStatus, errorIndex, varBindTable, cbCtx):
-            (self, varBindHead, varBindTotalTable, appReturn) = cbCtx
-            if errorStatus or \
-               errorIndication and not self.ignoreNonIncreasingOid or \
-               errorIndication and self.ignoreNonIncreasingOid and \
-               not isinstance(errorIndication, errind.OidNotIncreasing):
-                appReturn['errorIndication'] = errorIndication
-                appReturn['errorStatus'] = errorStatus
-                appReturn['errorIndex'] = errorIndex
-                appReturn['varBindTable'] = varBindTable
-                return
-            else:
-                while varBindTable:
-                    if len(varBindTable[-1]) != len(varBindHead):
-                        # Fix possibly non-rectangular table
-                        del varBindTable[-1]
-                    else:
-                        break
-                    
-                varBindTotalTable.extend(varBindTable) # XXX out of table 
-                                                       # rows possible
-                varBindTableRow = varBindTable[-1]
-                for idx in range(len(varBindTableRow)):
-                    name, val = varBindTableRow[idx]
-                    if not isinstance(val, univ.Null):
-                        if self.lexicographicMode:
+                        if lexicographicMode or \
+                               hasattr(self, 'lexicographicMode') and \
+                               self.lexicographicMode:  # obsolete
                             if varBindHead[idx] <= name:
                                 break
                         else:
@@ -511,27 +436,121 @@ class CommandGenerator:
                     appReturn['varBindTable'] = varBindTotalTable
                     return
                 
-            return 1 # continue table retrieval
+                varBindTotalTable.extend(varBindTable)
 
-        varBindHead = []
-        for varName in varNames:
-            name, suffix = mibvar.mibNameToOid(
-                self.__asynCmdGen.mibViewController, varName
-                )
-            varBindHead.append(univ.ObjectIdentifier(name + suffix))
+                return 1 # continue table retrieval
+
+        lookupNames = kwargs.get('lookupNames', False)        
+        lookupValues = kwargs.get('lookupValues', False)
+        lexicographicMode = kwargs.get('lexicographicMode', False)
+        maxRows = kwargs.get('maxRows', 0)
+        ignoreNonIncreasingOid = kwargs.get('ignoreNonIncreasingOid', False)
+
+        varBindHead = [ univ.ObjectIdentifier(x[0]) for x in self.__asynCmdGen.makeReadVarBinds(varNames) ]
 
         appReturn = {}
-        
-        self.__asynCmdGen.bulkCmd(
-            authData, transportTarget, nonRepeaters, maxRepetitions,
-            varNames, (__cbFun, (self, varBindHead, [], appReturn))
-            )
+        self.__asynCmdGen.nextCmd(
+            authData,
+            transportTarget,
+            varNames,
+            (__cbFun, (self, varBindHead, [], appReturn)),
+            lookupNames, lookupValues
+        )
 
         self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
-        
+
         return (
             appReturn['errorIndication'],
             appReturn['errorStatus'],
             appReturn['errorIndex'],
             appReturn['varBindTable']
-            )
+        )
+
+    def bulkCmd(self, authData, transportTarget,
+                nonRepeaters, maxRepetitions, *varNames, **kwargs):
+        def __cbFun(sendRequestHandle, errorIndication,
+                    errorStatus, errorIndex, varBindTable, cbCtx):
+            (self, varBindHead, varBindTotalTable, appReturn) = cbCtx
+            if (ignoreNonIncreasingOid or \
+                        hasattr(self, 'ignoreNonIncreasingOid') and \
+                        self.ignoreNonIncreasingOid ) and \
+                    errorIndication and \
+                    isinstance(errorIndication, errind.OidNotIncreasing):
+                errorIndication = None
+            if errorStatus or errorIndication:
+                appReturn['errorIndication'] = errorIndication
+                appReturn['errorStatus'] = errorStatus
+                appReturn['errorIndex'] = errorIndex
+                appReturn['varBindTable'] = varBindTable
+                return
+            else:
+                while varBindTable:
+                    if len(varBindTable[-1]) != len(varBindHead):
+                        # Fix possibly non-rectangular table
+                        del varBindTable[-1]
+                    else:
+                        break
+
+                varBindTotalTable.extend(varBindTable) # XXX out of table 
+                                                       # rows possible
+
+                if maxRows and len(varBindTotalTable) >= maxRows or \
+                        hasattr(self, 'maxRows') and self.maxRows and \
+                        len(varBindTotalTable) >= self.maxRows:  # obsolete
+                    appReturn['errorIndication'] = errorIndication
+                    appReturn['errorStatus'] = errorStatus
+                    appReturn['errorIndex'] = errorIndex
+                    if hasattr(self, 'maxRows'):
+                        appReturn['varBindTable'] = varBindTotalTable[:self.maxRows]
+                    else:
+                        appReturn['varBindTable'] = varBindTotalTable[:maxRows]
+                    return
+
+                varBindTableRow = varBindTable and varBindTable[-1] or varBindTable
+                for idx in range(len(varBindTableRow)):
+                    name, val = varBindTableRow[idx]
+                    if not isinstance(val, univ.Null):
+                        if lexicographicMode or \
+                               hasattr(self, 'lexicographicMode') and \
+                               self.lexicographicMode:  # obsolete
+                            if varBindHead[idx] <= name:
+                                break
+                        else:
+                            if varBindHead[idx].isPrefixOf(name):
+                                break
+                else:
+                    appReturn['errorIndication'] = errorIndication
+                    appReturn['errorStatus'] = errorStatus
+                    appReturn['errorIndex'] = errorIndex
+                    appReturn['varBindTable'] = varBindTotalTable
+                    return
+
+                return 1 # continue table retrieval
+
+        lookupNames = kwargs.get('lookupNames', False)        
+        lookupValues = kwargs.get('lookupValues', False)
+        lexicographicMode = kwargs.get('lexicographicMode', False)
+        maxRows = kwargs.get('maxRows', 0)
+        ignoreNonIncreasingOid = kwargs.get('ignoreNonIncreasingOid', False)
+
+        varBindHead = [ univ.ObjectIdentifier(x[0]) for x in self.__asynCmdGen.makeReadVarBinds(varNames) ]
+
+        appReturn = {}
+        
+        self.__asynCmdGen.bulkCmd(
+            authData,
+            transportTarget,
+            nonRepeaters, maxRepetitions,
+            varNames,
+            (__cbFun, (self, varBindHead, [], appReturn)),
+            lookupNames, lookupValues
+        )
+
+        self.__asynCmdGen.snmpEngine.transportDispatcher.runDispatcher()
+
+        return (
+            appReturn['errorIndication'],
+            appReturn['errorStatus'],
+            appReturn['errorIndex'],
+            appReturn['varBindTable']
+        )
