@@ -1,5 +1,5 @@
 """SNMP v3 Message Processing and Dispatching (RFC3412)"""
-import time, sys
+import sys
 from pyasn1.compat.octets import null
 from pysnmp.smi import builder, instrum
 from pysnmp.proto import errind, error, cache
@@ -56,13 +56,13 @@ class MsgAndPduDispatcher:
             k = (contextEngineId, pduType)
             if k in self.__appsRegistration:
                 raise error.ProtocolError(
-                    'Duplicate registration %s/%s' % (contextEngineId, pduType)
+                    'Duplicate registration %r/%s' % (contextEngineId, pduType)
                     )
 
             # 4.3.4
             self.__appsRegistration[k] = processPdu
 
-        debug.logger & debug.flagDsp and debug.logger('registerContextEngineId: contextEngineId %s pduTypes %s' % (repr(contextEngineId), pduTypes))
+        debug.logger & debug.flagDsp and debug.logger('registerContextEngineId: contextEngineId %r pduTypes %s' % (contextEngineId, pduTypes))
     # 4.4.1
     def unregisterContextEngineId(self, contextEngineId, pduTypes):
         """Unregister application with dispatcher"""
@@ -76,7 +76,7 @@ class MsgAndPduDispatcher:
             if k in self.__appsRegistration:
                 del self.__appsRegistration[k]
 
-        debug.logger & debug.flagDsp and debug.logger('unregisterContextEngineId: contextEngineId %s pduTypes %s' % (repr(contextEngineId), pduTypes))
+        debug.logger & debug.flagDsp and debug.logger('unregisterContextEngineId: contextEngineId %r pduTypes %s' % (contextEngineId, pduTypes))
 
     def getRegisteredApp(self, contextEngineId, pduType):
         k = (contextEngineId, pduType)
@@ -104,7 +104,7 @@ class MsgAndPduDispatcher:
         pduVersion,
         PDU,
         expectResponse,
-        timeout=0,    # response items
+        timeout=0,    # timeout expressed in dispatcher ticks
         cbFun=None,
         cbCtx=None
         ):
@@ -127,10 +127,11 @@ class MsgAndPduDispatcher:
                 sendPduHandle,
                 messageProcessingModel=messageProcessingModel,
                 sendPduHandle=sendPduHandle,
-                timeout=timeout,
+                timeout=timeout+snmpEngine.transportDispatcher.getTimerTicks(),
                 cbFun=cbFun,
                 cbCtx=cbCtx
                 )
+            debug.logger & debug.flagDsp and debug.logger('sendPdu: current time in ticks %d' % (snmpEngine.transportDispatcher.getTimerTicks(),))
 
         debug.logger & debug.flagDsp and debug.logger('sendPdu: new sendPduHandle %s, timeout %s, cbFun %s' % (sendPduHandle, timeout, cbFun))
 
@@ -268,9 +269,9 @@ class MsgAndPduDispatcher:
         try:
             restOfWholeMsg = null # XXX fix decoder non-recursive return
             msgVersion = verdec.decodeMessageVersion(wholeMsg)
-        except PySnmpError:
-            snmpInAsn1ParseErrs, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInAsn1ParseErrs')
-            snmpInAsn1ParseErrs.syntax = snmpInAsn1ParseErrs.syntax + 1
+        except error.ProtocolError:
+            snmpInASNParseErrs, = self.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInASNParseErrs')
+            snmpInASNParseErrs.syntax = snmpInASNParseErrs.syntax + 1
             return null  # n.b the whole buffer gets dropped
 
         debug.logger & debug.flagDsp and debug.logger('receiveMessage: msgVersion %s, msg decoded' % msgVersion)
@@ -313,7 +314,7 @@ class MsgAndPduDispatcher:
             if 'sendPduHandle' in statusInformation:
                 # Dropped REPORT -- re-run pending reqs queue as some
                 # of them may be waiting for this REPORT
-                debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP failed, statusInformation %s' % statusInformation)
+                debug.logger & debug.flagDsp and debug.logger('receiveMessage: MP failed, statusInformation %s, forcing a retry' % statusInformation)
                 self.__expireRequest(
                     statusInformation['sendPduHandle'],
                     self.__cache.pop(statusInformation['sendPduHandle']),
@@ -456,8 +457,10 @@ class MsgAndPduDispatcher:
 
     def __expireRequest(self, cacheKey, cachedParams, snmpEngine,
                         statusInformation=None):
+        timeNow = snmpEngine.transportDispatcher.getTimerTicks()
         timeoutAt = cachedParams['timeout']
-        if statusInformation is None and time.time() < timeoutAt:
+
+        if statusInformation is None and timeNow < timeoutAt:
             return
 
         processResponsePdu = cachedParams['cbFun']
