@@ -40,9 +40,9 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         aes256.Aes256.serviceID: aes256.Aes256(),
         nopriv.NoPriv.serviceID: nopriv.NoPriv()
         }
-    _securityParametersSpec = UsmSecurityParameters()
     def __init__(self):
         AbstractSecurityModel.__init__(self)
+        self.__securityParametersSpec = UsmSecurityParameters()
         self.__timeline = {}
         self.__timelineExpQueue = {}
         self.__expirationTimer = 0
@@ -203,7 +203,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             if 'usmUserAuthProtocol' in cachedSecurityData:
                 usmUserAuthProtocol = cachedSecurityData['usmUserAuthProtocol']
             else:
-                usmUserAuthProtocol = None
+                usmUserAuthProtocol = noauth.NoAuth.serviceID
             if 'usmUserAuthKeyLocalized' in cachedSecurityData:
                 usmUserAuthKeyLocalized = cachedSecurityData['usmUserAuthKeyLocalized']
             else:
@@ -211,7 +211,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             if 'usmUserPrivProtocol' in cachedSecurityData:
                 usmUserPrivProtocol = cachedSecurityData['usmUserPrivProtocol']
             else:
-                usmUserPrivProtocol = None
+                usmUserPrivProtocol = nopriv.NoPriv.serviceID
             if 'usmUserPrivKeyLocalized' in cachedSecurityData:
                 usmUserPrivKeyLocalized = cachedSecurityData['usmUserPrivKeyLocalized']
             else:
@@ -255,29 +255,31 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         else:
             # empty username used for engineID discovery
             usmUserName = usmUserSecurityName = null
-            usmUserAuthProtocol = usmUserAuthKeyLocalized = None
-            usmUserPrivProtocol = usmUserPrivKeyLocalized = None
+            usmUserAuthProtocol = noauth.NoAuth.serviceID
+            usmUserPrivProtocol = nopriv.NoPriv.serviceID
+            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use empty USM data')
             
-        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: local user usmUserName %s usmUserAuthProtocol %s usmUserPrivProtocol %s securityEngineID %s securityName %s' % (usmUserName, usmUserAuthProtocol, usmUserPrivProtocol, repr(securityEngineID), securityName))
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: local user usmUserName %r usmUserAuthProtocol %s usmUserPrivProtocol %s securityEngineID %r securityName %r' % (usmUserName, usmUserAuthProtocol, usmUserPrivProtocol, securityEngineID, securityName))
 
         msg = globalData
         
         # 3.1.2
         if securityLevel == 3:
-            if not usmUserAuthProtocol or not usmUserPrivProtocol:
+            if usmUserAuthProtocol == noauth.NoAuth.serviceID or \
+               usmUserPrivProtocol == nopriv.NoPriv.serviceID:
                 raise error.StatusInformation(
                     errorIndication = errind.unsupportedSecurityLevel
                     )
 
         # 3.1.3
         if securityLevel == 3 or securityLevel == 2:
-            if not usmUserAuthProtocol:
+            if usmUserAuthProtocol == noauth.NoAuth.serviceID:
                 raise error.StatusInformation(
                     errorIndication = errind.unsupportedSecurityLevel
                     )
 
-        securityParameters = self._securityParametersSpec
+        securityParameters = self.__securityParametersSpec
 
         scopedPDUData = msg.setComponentByPosition(3).getComponentByPosition(3)
         scopedPDUData.setComponentByPosition(
@@ -299,7 +301,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             else:
                 # 2.3 XXX is this correct?
                 snmpEngineBoots = snmpEngineTime = 0
-                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: no timeline for securityEngineID %s' % repr(securityEngineID))
+                debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: no timeline for securityEngineID %r' % (securityEngineID,))
         # 3.1.6.b
         elif securityStateReference is not None:  # XXX Report?
             ( snmpEngineBoots,
@@ -312,7 +314,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             snmpEngineBoots = snmpEngineTime = 0
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: assuming zero snmpEngineBoots, snmpEngineTime')
 
-        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use snmpEngineBoots %s snmpEngineTime %s for securityEngineID %s' % (snmpEngineBoots, snmpEngineTime, repr(securityEngineID)))
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: use snmpEngineBoots %s snmpEngineTime %s for securityEngineID %r' % (snmpEngineBoots, snmpEngineTime, securityEngineID))
 
         # 3.1.4a
         if securityLevel == 3:
@@ -327,7 +329,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
 
             dataToEncrypt = encoder.encode(scopedPDU)
             
-            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU encoded into %s' % repr(dataToEncrypt))
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU encoded into %s' % debug.hexdump(dataToEncrypt))
 
             ( encryptedData,
               privParameters ) = privHandler.encryptData(
@@ -343,7 +345,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                 1, encryptedData, verifyConstraints=False
                 )
 
-            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU ciphered')
+            debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: scopedPDU ciphered into %s' % debug.hexdump(encryptedData))
 
         # 3.1.4b
         elif securityLevel == 1 or securityLevel == 2:
@@ -394,6 +396,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             authenticatedWholeMsg = authHandler.authenticateOutgoingMsg(
                 usmUserAuthKeyLocalized, wholeMsg
                 )
+
         # 3.1.8b
         else:
             securityParameters.setComponentByPosition(
@@ -409,6 +412,9 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: plain outgoing msg: %s' % msg.prettyPrint())
 
             authenticatedWholeMsg = encoder.encode(msg)
+
+        debug.logger & debug.flagSM and debug.logger('__generateRequestOrResponseMsg: %s outgoing msg: %s' % (securityLevel > 1 and "authenticated" or "plain", debug.hexdump(authenticatedWholeMsg)))
+
         # 3.1.9
         return (
             msg.getComponentByPosition(2),
@@ -481,20 +487,23 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         # 3.2.9 -- moved up here to be able to report
         # maxSizeResponseScopedPDU on error
         # (48 - maximum SNMPv3 header length)
-        maxSizeResponseScopedPDU = maxMessageSize - len(securityParameters)-48
+        maxSizeResponseScopedPDU = int(maxMessageSize) - len(securityParameters) - 48
+
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: securityParameters %s' % debug.hexdump(securityParameters))
 
         # 3.2.1 
         try:
             securityParameters, rest = decoder.decode(
                 securityParameters,
-                asn1Spec=self._securityParametersSpec
+                asn1Spec=self.__securityParametersSpec
                 )
         except PyAsn1Error:
-           snmpInASNParseErrs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInASNParseErrs')
-           snmpInASNParseErrs.syntax = snmpInASNParseErrs.syntax + 1
-           raise error.StatusInformation(
-               errorIndication=errind.parseError
-               )
+            debug.logger & debug.flagSM and debug.logger('processIncomingMsg: %s' % (sys.exc_info()[1],))
+            snmpInASNParseErrs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMPv2-MIB', 'snmpInASNParseErrs')
+            snmpInASNParseErrs.syntax = snmpInASNParseErrs.syntax + 1
+            raise error.StatusInformation(
+                errorIndication=errind.parseError
+                )
 
         debug.logger & debug.flagSM and debug.logger('processIncomingMsg: %s' % (securityParameters.prettyPrint(),))
 
@@ -514,7 +523,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
 
         # 3.2.3
         if msgAuthoritativeEngineID not in self.__timeline:
-            debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %s' % repr(msgAuthoritativeEngineID))
+            debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %r' % (msgAuthoritativeEngineID,))
             if not msgAuthoritativeEngineID:
                 # 3.2.3b
                 usmStatsUnknownEngineIDs, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownEngineIDs')
@@ -558,7 +567,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
  
         msgUserName = securityParameters.getComponentByPosition(3)
 
-        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read from securityParams msgAuthoritativeEngineID %s msgUserName %s' % (repr(msgAuthoritativeEngineID), msgUserName))
+        debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read from securityParams msgAuthoritativeEngineID %r msgUserName %r' % (msgAuthoritativeEngineID, msgUserName))
         
         if msgUserName:
             # 3.2.4
@@ -590,7 +599,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                         debug.logger & debug.flagSM and debug.logger('processIncomingMsg: cloned user info')
                     except NoSuchInstanceError:
                         __reportUnknownName = 1
-                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %s msgUserName %s' % (repr(msgAuthoritativeEngineID), msgUserName))
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: unknown securityEngineID %r msgUserName %r' % (msgAuthoritativeEngineID, msgUserName))
                 if __reportUnknownName:
                         usmStatsUnknownUserNames, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnknownUserNames')
                         usmStatsUnknownUserNames.syntax = usmStatsUnknownUserNames.syntax+1
@@ -607,8 +616,9 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
         else:
             # empty username used for engineID discovery
             usmUserName = usmUserSecurityName = null
-            usmUserAuthProtocol = usmUserAuthKeyLocalized = None
-            usmUserPrivProtocol = usmUserPrivKeyLocalized = None
+            usmUserAuthProtocol = noauth.NoAuth.serviceID
+            usmUserPrivProtocol = nopriv.NoPriv.serviceID
+            usmUserAuthKeyLocalized = usmUserPrivKeyLocalized = None
 
         debug.logger & debug.flagSM and debug.logger('processIncomingMsg: now have usmUserSecurityName %s usmUserAuthProtocol %s usmUserPrivProtocol %s for msgUserName %s' % (usmUserSecurityName, usmUserAuthProtocol, usmUserPrivProtocol, msgUserName))
 
@@ -623,26 +633,39 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             )
 
         # 3.2.5
-        __reportError = 0
-        if securityLevel == 3:
-            if not usmUserAuthProtocol or not usmUserPrivProtocol:
-                __reportError = 1
+        if msgAuthoritativeEngineID == snmpEngineID:
+            # Authoritative SNMP engine: make sure securityLevel is sufficient
+            __badSecIndication = None
+            if securityLevel == 3:
+                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
+                    __badSecIndication = 'authPriv wanted while auth not expected'
+                if usmUserPrivProtocol == nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'authPriv wanted while priv not expected'
             elif securityLevel == 2:
-                if not usmUserAuthProtocol:
-                    __reportError = 1
-        if __reportError:
-            usmStatsUnsupportedSecLevels, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnsupportedSecLevels')
-            usmStatsUnsupportedSecLevels.syntax = usmStatsUnsupportedSecLevels.syntax + 1
-            raise error.StatusInformation(
-                errorIndication=errind.unsupportedSecurityLevel,
-                oid=usmStatsUnknownEngineIDs.name,
-                val=usmStatsUnknownEngineIDs.syntax,
-                securityStateReference=securityStateReference,
-                securityLevel=securityLevel,
-                contextEngineId=contextEngineId,
-                contextName=contextName,
-                maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
-                )
+                if usmUserAuthProtocol == noauth.NoAuth.serviceID:
+                    __badSecIndication = 'authNoPriv wanted while auth not expected'
+                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'authNoPriv wanted while priv expected'
+
+            elif securityLevel == 1:
+                if usmUserAuthProtocol != noauth.NoAuth.serviceID:
+                    __badSecIndication = 'noAuthNoPriv wanted while auth expected'
+                if usmUserPrivProtocol != nopriv.NoPriv.serviceID:
+                    __badSecIndication = 'noAuthNoPriv wanted while priv expected'
+            if __badSecIndication:
+                usmStatsUnsupportedSecLevels, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsUnsupportedSecLevels')
+                usmStatsUnsupportedSecLevels.syntax = usmStatsUnsupportedSecLevels.syntax + 1
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: reporting inappropriate security level for user %s: %s' % (msgUserName, __badSecIndication))
+                raise error.StatusInformation(
+                    errorIndication=errind.unsupportedSecurityLevel,
+                    oid=usmStatsUnsupportedSecLevels.name,
+                    val=usmStatsUnsupportedSecLevels.syntax,
+                    securityStateReference=securityStateReference,
+                    securityLevel=securityLevel,
+                    contextEngineId=contextEngineId,
+                    contextName=contextName,
+                    maxSizeResponseScopedPDU=maxSizeResponseScopedPDU
+                    )
 
         # 3.2.6
         if securityLevel == 3 or securityLevel == 2:
@@ -683,14 +706,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                     int(time.time())
                     )
                 
-                expireAt = self.__expirationTimer + 300
+                expireAt = int(self.__expirationTimer + 300 / snmpEngine.transportDispatcher.getTimerResolution())
                 if expireAt not in self.__timelineExpQueue:
                     self.__timelineExpQueue[expireAt] = []
                 self.__timelineExpQueue[expireAt].append(
                     msgAuthoritativeEngineID
                     )
                     
-                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: store timeline for securityEngineID %s' % repr(msgAuthoritativeEngineID))
+                debug.logger & debug.flagSM and debug.logger('processIncomingMsg: store timeline for securityEngineID %r' % (msgAuthoritativeEngineID,))
             
         # 3.2.7
         if securityLevel == 3 or securityLevel == 2:
@@ -713,7 +736,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                         ]
                     # time passed since last talk with this SNMP engine
                     idleTime = int(time.time())-latestUpdateTimestamp
-                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read timeline snmpEngineBoots %s snmpEngineTime %s for msgAuthoritativeEngineID %s, idle time %s secs' % (snmpEngineBoots, snmpEngineTime, repr(msgAuthoritativeEngineID), idleTime))
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: read timeline snmpEngineBoots %s snmpEngineTime %s for msgAuthoritativeEngineID %r, idle time %s secs' % (snmpEngineBoots, snmpEngineTime, msgAuthoritativeEngineID, idleTime))
                 else:
                     raise error.ProtocolError('Peer SNMP engine info missing')
 
@@ -750,14 +773,14 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                         msgAuthoritativeEngineTime,
                         int(time.time())
                         )
-                    expireAt = self.__expirationTimer + 300
+                    expireAt = int(self.__expirationTimer + 300 / snmpEngine.transportDispatcher.getTimerResolution())
                     if expireAt not in self.__timelineExpQueue:
                         self.__timelineExpQueue[expireAt] = []
                     self.__timelineExpQueue[expireAt].append(
                         msgAuthoritativeEngineID
                         )
 
-                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: stored timeline msgAuthoritativeEngineBoots %s msgAuthoritativeEngineTime %s for msgAuthoritativeEngineID %s' % (msgAuthoritativeEngineBoots, msgAuthoritativeEngineTime, repr(msgAuthoritativeEngineID)))
+                    debug.logger & debug.flagSM and debug.logger('processIncomingMsg: stored timeline msgAuthoritativeEngineBoots %s msgAuthoritativeEngineTime %s for msgAuthoritativeEngineID %r' % (msgAuthoritativeEngineBoots, msgAuthoritativeEngineTime, msgAuthoritativeEngineID))
                     
                 # 3.2.7b.2
                 if snmpEngineBoots == 2147483647 or \
@@ -791,7 +814,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
                       securityParameters.getComponentByPosition(5) ),
                     encryptedPDU
                     )
-               debug.logger & debug.flagSM and debug.logger('processIncomingMsg: PDU deciphered into %s' % repr(decryptedData))
+               debug.logger & debug.flagSM and debug.logger('processIncomingMsg: PDU deciphered into %s' % debug.hexdump(decryptedData))
             except error.StatusInformation:
                 usmStatsDecryptionErrors, = snmpEngine.msgAndPduDsp.mibInstrumController.mibBuilder.importSymbols('__SNMP-USER-BASED-SM-MIB', 'usmStatsDecryptionErrors')
                 usmStatsDecryptionErrors.syntax = usmStatsDecryptionErrors.syntax+1
@@ -859,7 +882,7 @@ class SnmpUSMSecurityModel(AbstractSecurityModel):
             for engineIdKey in self.__timelineExpQueue[self.__expirationTimer]:
                 if engineIdKey in self.__timeline:
                     del self.__timeline[engineIdKey]
-                    debug.logger & debug.flagSM and debug.logger('__expireEnginesInfo: expiring %s' % (engineIdKey,))
+                    debug.logger & debug.flagSM and debug.logger('__expireTimelineInfo: expiring %r' % (engineIdKey,))
             del self.__timelineExpQueue[self.__expirationTimer]
         self.__expirationTimer = self.__expirationTimer + 1
         
